-- rsiTurtle.lua
-- Created on:  2020-10-16
-- Author:      Rétro
--
-- Rétro Stonks Industries - Tout droits réservés
--
-- API pour tortues RSI

---
-- \class Tortue RSI
local RSI_Turtle = {}
RSI_Turtle.__index = RSI_Turtle

require("printf")

---
-- \brief Affiche un hello world
function RSI_Turtle:helloWorld()
    print("Hello World!")
end

---
-- \brief Constructeur de RSI_Turtle
function RSI_Turtle:init()
    local turtle = {}
    setmetatable(turtle,RSI_Turtle)
    
    ---
    -- \brief Dictionnaire pour la position (x,y,z).
    --        Le système de coordonnées est celui de minecraft:
    --          x: Avant
    --          y: Altitude
    --          z: Droite
    turtle.pos_ = {}
    turtle.pos_["x"] = 0
    turtle.pos_["y"] = 0
    turtle.pos_["z"] = 0
    
    ---
    -- \brief Orientation de la tortue en cadrant horloge:
    --          0:  Avant
    --          3:  Droite
    --          6:  Arrière
    --          9:  Gauche
    turtle.facing_ = 0
    
    return turtle
end


---
-- \brief Affiche la position et l'orientation de la tortue
function RSI_Turtle:print()
    printf("Position:\n X=%d\n Y=%d\n Z=%d\n", self.pos_["x"],
                    self.pos_["y"], self.pos_["z"])
    printf("Orientation: ")

    if( self.facing_ == 3 )
    then
        print("Droite")
    elseif ( self.facing_ == 6 )
    then
        print("Arriere")
    elseif ( self.facing_ == 9 )
    then
        print("Gauche")
    elseif ( self.facing_ == 0 )
    then
        print("Avant")
    end
end


---
-- \brief   Affiche la position et l'orientation de la tortue
-- \return int,int,int  Positions X, Y et Z
function RSI_Turtle:getPos()
    return self.pos_["x"], self.pos_["y"], self.pos_["z"]
end


---
-- \brief Déplacement parallèle au sol
-- \return bool Succès
function RSI_Turtle:moveTo2D(x, z)
    return self:move2D(x - self.pos_["x"], z - self.pos_["z"])
end


---
-- \brief Déplacement dans les trois dimensions
-- \return bool Succès
function RSI_Turtle:moveTo(x, y, z)
    return self:move(x - self.pos_["x"], y - self.pos_["y"], z - self.pos_["z"])
end


---
-- \brief Retourner au point d'origine (0,0,0)
-- \return bool Succès
function RSI_Turtle:moveToOrigin()
    return self:moveTo(0,0,0)
end


---
-- \brief Déplacement parallèle au sol
--        Ordre de déplacement: X, Z
-- \return bool Succès
function RSI_Turtle:move2D(dx, dz)
    local success = true

    if dx >= 0
    then
        self:face(12)
        success = self:moveForward(dx)
    elseif dx < 0
    then
        self:face(6)
        success = self:moveForward(-1*dx)
    end

    if not success
    then
        return false
    end
    
    if (self.facing_ == 0)
    then
        if dz < 0
        then
            self:turnLeft()
        elseif dz > 0
        then
            self:turnRight()
        end
    elseif (self.facing_ == 6)
    then
        if dz > 0
        then
            self:turnLeft()
        elseif dz < 0
        then
            self:turnRight()
        end
    end
    
    if not success
    then
        return false
    end
    succes = self:moveForward(math.abs(dz))


    return success
end


---
-- \brief Déplacement dans les trois dimensions
--        Ordre de déplacement: X, Z, Y
-- \return bool Succès
function RSI_Turtle:move(dx, dy, dz)
    local success = true

    if (dy > 0)
    then
        for i=1,dy,1
        do
            success = self:up()
        end
    elseif dy < 0
    then
        for i=-1,dy,-1
        do
            success = self:down()
        end
    end

    if (success) then
        return self:move2D(dx, dz)
    else
        return false
    end
end


---
-- \brief Déplacement s'assurant de rester au niveau du sol
-- escalade et descend automatiquement les pentes
function RSI_Turtle:moveToGrounded(x, z)
    --TODO: Implementer
end


---
-- \brief Déplacement s'assurant de rester au niveau du sol
-- escalade et descend automatiquement les pentes
-- \return bool Succès
function RSI_Turtle:moveGrounded(dx, dz)
    --TODO: Implementer
end


---
-- \brief Tourne la tortue face à la direction voulue.
-- \param Orientation en heures (3, 6, 9, 12)
-- \return bool Succès
function RSI_Turtle:face(orientation)
    if (orientation ~= 3 and orientation ~= 6 and 
        orientation ~= 9 and orientation ~= 12)
    then
        return false
    end

    local diff = (orientation - self.facing_) % 12
    if diff == 9 then
        self:turnLeft()
    elseif diff == 3 then
        self:turnRight()
    elseif diff == 6 then
        self:turnLeft()
        self:turnLeft()
    end
    return true
end

---
-- \brief Déplacement avant
-- \return bool Succès
function RSI_Turtle:forward()
    return self:moveForward(1)
end

---
-- \brief Déplacement avant multiple
-- \param distance Distance à parcourir
-- \return bool Succès
function RSI_Turtle:moveForward(distance)
    local traveled = 0

    while (traveled < distance) and turtle.forward()
    do
        traveled = traveled + 1
    end

    if self.facing_ == 0 then
        self.pos_["x"] = self.pos_["x"] + traveled
    elseif self.facing_ == 3 then
        self.pos_["z"] = self.pos_["z"] + traveled 
    elseif self.facing_ == 6 then
        self.pos_["x"] = self.pos_["x"] - traveled
    elseif self.facing_ == 9 then
        self.pos_["z"] = self.pos_["z"] - traveled
    end

    return (traveled == distance)
end


---
-- \brief Déplacement arrière
-- \return bool Succès
function RSI_Turtle:backward()
    return self:moveBackward(1)
end


---
-- \brief Déplacement arrière
-- \param distance Distance à parcourir
-- \return bool Succès
function RSI_Turtle:moveBackward(distance)
    local traveled = 0

    while (traveled < distance) and turtle.back()
    do
        traveled = traveled + 1
    end
    
    if result then
        if self.facing_ == 0 then
            self.pos_["x"] = self.pos_["x"] - traveled
        elseif self.facing_ == 3 then
            self.pos_["z"] = self.pos_["z"] - traveled
        elseif self.facing_ == 6 then
            self.pos_["x"] = self.pos_["x"] + traveled
        elseif self.facing_ == 9 then
            self.pos_["z"] = self.pos_["z"] + traveled
        end
    end

    return (traveled == distance)
end


---
-- \brief Déplacement haut
-- \return bool Succès
function RSI_Turtle:up()
    result = turtle.up()
    
    if result then
        self.pos_["y"] = self.pos_["y"] + 1
    end
    return result
end


---
-- \brief Déplacement bas
-- \return bool Succès
function RSI_Turtle:down()
    result = turtle.down()
    
    if result then
        self.pos_["y"] = self.pos_["y"] - 1
    end
    return result
end


---
-- \brief Tourne 90 degrés dans le sens contraire des aiguilles d'une montre
function RSI_Turtle:turnLeft()
    turtle.turnLeft()
    self.facing_ = (self.facing_ - 3) % 12
end


---
-- \brief Tourne 90 degrés dans le sens des aiguilles d'une montre
function RSI_Turtle:turnRight()
    turtle.turnRight()
    self.facing_ = (self.facing_ + 3) % 12
end


---
-- \brief Déplacement à gauche
-- \return bool Succès
function RSI_Turtle:left()
    self:turnLeft()
    result = self:forward()
    self:turnRight()
    return result
end


---
-- \brief Déplacement à droite
-- \return bool Succès
function RSI_Turtle:right()
    self:turnRight()
    result = self:forward()
    self:turnLeft()
    return result
end


---
-- \brief Descendre jusqu'à rencontrer le sol
-- \return bool Succès si a descendu d'au moins un bloc
function RSI_Turtle:ground()
    success = self:down()
    while self:down() do end
    return success
end


---
-- \brief Quadriller une zone carrée. La tortue traverse chaque case, ligne par 
--          ligne. N'affecte pas l'altitude.
-- \param pointeur_de_fonction action Pointeur de l'action à réaliser juste
--          après être arrivé sur chaque nouveau bloc.
-- \return bool Succès
function RSI_Turtle:goThroughArea(action)
    --TODO: Implementer
end


---
-- \brief Quadriller une zone carrée. La tortue traverse chaque case, ligne par 
--          ligne. Reste au niveau du sol.
-- \param pointeur_de_fonction action Pointeur de l'action à réaliser juste
--          après être arrivé sur chaque nouveau bloc.
-- \return bool Succès
function RSI_Turtle:goThroughAreaGround(action)
    --TODO: Implementer
end


---
-- \brief Vérifier l'ID du bloc en face
-- \param Nom du bloc à vérifier
-- \return bool Bloc en face identique à l'argument
function RSI_Turtle:facesBlock(block_id)
    --TODO: Implementer
end

return RSI_Turtle
